# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import os

from . import config_iterators
from . import config_splitters
from . import parallel_loop
from .abstract_dd import AbstractDD
from .abstract_parallel_dd import AbstractParallelDD

logger = logging.getLogger(__name__)


class CombinedParallelDD(AbstractParallelDD):

    def __init__(self, test, *, cache=None, split=config_splitters.zeller,
                 proc_num=os.cpu_count(), max_utilization=100,
                 config_iterator=config_iterators.forward):
        """
        Initialize a CombinedParallelDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param split: Splitter method to break a configuration up to n part.
        :param proc_num: The level of parallelization.
        :param max_utilization: The maximum CPU utilization accepted.
        :param config_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        """
        AbstractParallelDD.__init__(self, test, split, proc_num, max_utilization, cache=cache)

        self._config_iterator = config_iterator

    def _dd(self, config):
        """
        Calculates a 1-minimal subset of the config that is still interesting using multiple processes.
        Subset and complement based blocks are mixed and don't wait for each other.

        :param config: The input configuration.
        :return: A minimal subset of the current configuration what is still interesting (if any).
        """
        n = 2
        run = 1
        complement_offset = 0

        while True:
            assert self.test(config, 'assert') == self.FAIL

            subsets = self._split(config, n)

            logger.info('Run #%d: trying %s.', run, ' + '.join([str(len(subsets[i])) for i in range(n)]))

            # Reset fail index.
            self._fail_index.value = -1

            ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
            for i in self._config_iterator(2 * n):
                if i is None:
                    continue

                if i < n:
                    config_id = AbstractDD.config_id(run, 's', i)
                    config_set = subsets[i]
                else:
                    j = int((i - n + complement_offset) % n)
                    config_id = AbstractDD.config_id(run, 'c', j)
                    config_set = self.minus(config, subsets[j])

                # If we checked this test before, return its result
                outcome = self.lookup_cache(config_set, config_id)
                if outcome == self.PASS:
                    continue
                elif outcome == self.FAIL:
                    self._fail_index.value = i
                    break

                # Break if we found a FAIL either in the cache or be testing it now.
                if not ploop.do(self._loop_body, (config_set, i, config_id)):
                    # if do() returned False, the test was not started
                    break
            ploop.join()

            # fvalue contains the index of the cycle in the previous loop
            # which was found interesting. Otherwise it's -1.
            fvalue = self._fail_index.value
            if fvalue != -1:
                # Subset fail.
                if fvalue < n:
                    logger.info('Reduced to %d units.', len(subsets[fvalue]))
                    logger.debug('New config: %r.', subsets[fvalue])

                    next_config = subsets[fvalue]
                    next_n = 2
                    complement_offset = 0
                # Complement fail.
                else:
                    j = int((fvalue - n + complement_offset) % n)
                    complement = self.minus(config, subsets[j])
                    logger.info('Reduced to %d units.', len(complement))
                    logger.debug('New config: %r.', complement)

                    next_config = complement
                    next_n = max(n - 1, 2)

                    # In next run, start removing the following subset.
                    complement_offset = j
            else:
                next_config = config
                next_n = min(len(config), n * 2)
                logger.info('Increase granularity to %d.', next_n)
                complement_offset = (complement_offset * next_n) / n

            # Minimization ends if no interesting configuration was found by the finest splitting or
            # if the configuration is already reduced to a single unit.
            if fvalue == -1 and n == len(config):
                # No further minimizing.
                logger.info('Done.')
                return config

            if fvalue != -1 and len(next_config) == 1:
                # No further minimizing.
                logger.info('Done.')
                return next_config

            config = next_config
            n = next_n
            run += 1



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from .outcome_cache import OutcomeCache

logger = logging.getLogger(__name__)


class AbstractDD(object):
    """Abstract super-class of the parallel and non-parallel DD classes."""

    # Test outcomes.
    PASS = 'PASS'
    FAIL = 'FAIL'

    def __init__(self, test, split, *, cache=None):
        """
        Initialise an abstract DD class. Not to be called directly,
        only by super calls in subclass initializers.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n parts.
        :param cache: Cache object to use.
        """
        self._test = test
        self._split = split
        self._cache = cache or OutcomeCache()

    @staticmethod
    def config_id(run, dir, i):
        """
        Create identifier for the current task.

        :param run: The index of the current iteration.
        :param dir: The direction of reduce: either s(ubest) or c(omplement).
        :param i: The index of the current test in the iteration.
        :return: Config id in (run)_(dir)_(i) format.
        """
        return '%d_%s_%d' % (run, dir, i)

    @staticmethod
    def pretty_config_id(config_id):
        """
        Create beautified identifier for the current task.

        :param config_id: Config ID in form (run)_(dir)_(i).
        :return: Config ID in form (run) / (dir) / (i).
        """
        return config_id.replace('_', ' / ')

    def lookup_cache(self, config, config_id):
        """
        Perform a cache lookup if caching is enabled.

        :param config: The configuration we are looking for.
        :param config_id: The ID describing the configuration (only for debug message).
        :return: None if caching is disabled, PASS or FAIL otherwise.
        """
        cached_result = self._cache.lookup(config)
        if cached_result is not None:
            logger.debug('\t[ %s ]: cache = %r', AbstractDD.pretty_config_id(config_id), cached_result)

        return cached_result

    @staticmethod
    def minus(c1, c2):
        """Return a list of all elements of C1 that are not in C2."""
        c2 = set(c2)
        return [c for c in c1 if c not in c2]

    def _dd(self, config):
        """
        To be overridden by subclasses.

        :param config: The input configuration.
        :return: A minimal subset of the current configuration what is still interesting (if any).
        """
        pass

    def ddmin(self, config):
        """
        Return a 1-minimal failing subset of the initial configuration.

        :param config: The initial configuration that will be reduced.
        :return: 1-minimal failing configuration.
        """

        if len(config) < 2:
            assert self.test(config, 'assert') == self.FAIL
            logger.info('Test case is minimal already.')
            return config

        logger.debug('dd(%r) ...', config)

        outcome = self._dd(config)

        logger.debug('dd(%r) = %r', config, outcome)

        return outcome

    def test(self, config, config_id):
        """
        Test a single configuration and save the result in cache.

        :param config: The current configuration to test.
        :param config_id: Unique ID that will be used to save tests to easily identifiable directories.
        :return: PASS or FAIL
        """

        logger.debug('\t[ %s ]: test...', AbstractDD.pretty_config_id(config_id))

        outcome = self._test(config, config_id)

        logger.debug('\t[ %s ]: test = %r', AbstractDD.pretty_config_id(config_id), outcome)

        if config_id != 'assert':
            self._cache.add(config, outcome)

        return outcome



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import os

from . import config_iterators
from . import config_splitters
from . import parallel_loop
from .abstract_dd import AbstractDD
from .abstract_parallel_dd import AbstractParallelDD

logger = logging.getLogger(__name__)


class ParallelDD(AbstractParallelDD):

    def __init__(self, test, *, cache=None, split=config_splitters.zeller,
                 proc_num=os.cpu_count(), max_utilization=100,
                 subset_first=True, subset_iterator=config_iterators.forward, complement_iterator=config_iterators.forward):
        """
        Initialize a ParallelDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param split: Splitter method to break a configuration up to n part.
        :param proc_num: The level of parallelization.
        :param max_utilization: The maximum CPU utilization accepted.
        :param subset_first: Boolean value denoting whether the reduce has to start with the subset based approach or not.
        :param subset_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        """
        AbstractParallelDD.__init__(self, test, split, proc_num, max_utilization, cache=cache)

        self._subset_first = subset_first
        self._subset_iterator = subset_iterator
        self._complement_iterator = complement_iterator

    def _dd(self, config):
        """
        Calculates an 1-minimal subset of the config that is still interesting using multiple processes.
        Subset and complement based blocks are executed sequentially.

        :param config: The input configuration.
        :return: A minimal subset of the current configuration what is still interesting (if any).
        """
        n = 2
        run = 1
        complement_offset = 0

        if self._subset_first:
            first_test = self._test_subsets
            second_test = self._test_complements
        else:
            first_test = self._test_complements
            second_test = self._test_subsets

        while True:
            assert self.test(config, 'assert') == self.FAIL

            subsets = self._split(config, n)

            logger.info('Run #%d: trying %s.', run, ' + '.join([str(len(subsets[i])) for i in range(n)]))

            # Reset fail index.
            self._fail_index.value = -1

            next_config, next_n, complement_offset = first_test(n, run, config, subsets, complement_offset)
            if self._fail_index.value == -1:
                next_config, next_n, complement_offset = second_test(n, run, config, subsets, complement_offset)

            # If findex is still -1 then no interesting configuration was found in either loops.
            if self._fail_index.value == -1:
                next_config = config
                next_n = min(len(config), n * 2)
                logger.info('Increase granularity to %d.', next_n)
                complement_offset = (complement_offset * next_n) / n

            # Minimization ends if no interesting configuration was found by the finest splitting or
            # if the configuration is already reduced to a single unit.
            if self._fail_index.value == -1 and n == len(config):
                # No further minimizing.
                logger.info('Done.')
                return config

            if self._fail_index.value != -1 and len(next_config) == 1:
                # No further minimizing.
                logger.info('Done.')
                return next_config

            config = next_config
            n = next_n
            run += 1

    def _test_subsets(self, n, run, config, subsets, complement_offset):
        """
        Perform a subset based reduce task.

        :param n: The number of sets that the config is split to.
        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset need to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        # Looping through the subsets.
        ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
        for i in self._subset_iterator(n):
            if i is None:
                continue
            config_id = AbstractDD.config_id(run, 's', i)

            # If we had this test before, return the saved result.
            outcome = self.lookup_cache(subsets[i], config_id)
            if outcome == self.PASS:
                continue
            elif outcome == self.FAIL:
                self._fail_index.value = i
                break

            if not ploop.do(self._loop_body, (subsets[i], i, config_id)):
                # if do() returned False, the test was not started
                break
        ploop.join()

        fvalue = self._fail_index.value
        if fvalue != -1:
            logger.info('Reduced to %d units.', len(subsets[fvalue]))
            logger.debug('New config: %r.', subsets[fvalue])

            return subsets[fvalue], 2, 0

        return None, None, complement_offset

    def _test_complements(self, n, run, config, subsets, complement_offset):
        """
        Perform a complement based reduce task.

        :param n: The number of sets that the config is split to.
        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset need to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
        for j in self._complement_iterator(n):
            if j is None:
                continue

            i = int((j + complement_offset) % n)
            complement = self.minus(config, subsets[i])
            config_id = AbstractDD.config_id(run, 'c', i)

            # If we had this test before, return its result
            outcome = self.lookup_cache(complement, config_id)
            if outcome == self.PASS:
                continue
            elif outcome == self.FAIL:
                self._fail_index.value = i
                break

            if not ploop.do(self._loop_body, (complement, i, config_id)):
                # If do() returned False, the test was not started.
                break
        ploop.join()

        fvalue = self._fail_index.value
        if fvalue != -1:
            complement = self.minus(config, subsets[fvalue])
            logger.info('Reduced to %d units.', len(complement))
            logger.debug('New config: %r.', complement)

            # In next run, start removing the following subset.
            return complement, max(n - 1, 2), fvalue

        return None, None, complement_offset



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import inspect

from multiprocessing import Lock
from multiprocessing.managers import BaseManager

shared_manager_store = dict()


def shared_cache_decorator(cache_class):
    global shared_manager_store

    if cache_class in shared_manager_store:
        return shared_manager_store[cache_class].SharedCache

    class SharedDataManager(BaseManager):
        """Data manager to share the cache object between parallel processes."""
        pass

    class SharedCache(cache_class):
        """Thread-safe cache representation that stores the evaluated configurations and their outcome."""

        def __init__(self, *args, **kwargs):
            cache_class.__init__(self, *args, **kwargs)
            self._lock = Lock()

        def add(self, config, result):
            with self._lock:
                cache_class.add(self, config, result)

        def lookup(self, config):
            with self._lock:
                return cache_class.lookup(self, config)

        def clear(self):
            with self._lock:
                cache_class.clear(self)

    SharedDataManager.register('SharedCache', SharedCache, None,
                               [nv[0] for nv in inspect.getmembers(cache_class, lambda m: inspect.isfunction(m) and m.__name__ != '__init__')])
    getattr(SharedDataManager, 'SharedCache').__signature__ = inspect.signature(cache_class)
    data_manager = SharedDataManager()
    data_manager.start()
    shared_manager_store[cache_class] = data_manager

    return data_manager.SharedCache



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from . import config_iterators
from . import config_splitters
from .abstract_dd import AbstractDD
from .outcome_cache import ConfigCache

logger = logging.getLogger(__name__)


class LightDD(AbstractDD):
    """Single process version of the Delta Debugging algorithm."""

    def __init__(self, test, *, cache=None, split=config_splitters.zeller,
                 subset_first=True, subset_iterator=config_iterators.forward, complement_iterator=config_iterators.forward):
        """
        Initialize a LightDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param split: Splitter method to break a configuration up to n parts.
        :param subset_first: Boolean value denoting whether the reduce has to start with the subset based approach or not.
        :param subset_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        """
        cache = cache or ConfigCache()
        AbstractDD.__init__(self, test, split, cache=cache)

        self._subset_first = subset_first
        self._subset_iterator = subset_iterator
        self._complement_iterator = complement_iterator

    def _dd(self, config):
        """
        Calculates a 1-minimal subset of the config that is still interesting in single process mode.

        :param config: The input configuration.
        :return: A minimal subset of the current configuration what is still interesting (if any).
        """
        n = 2
        run = 1
        complement_offset = 0

        if self._subset_first:
            first_test = self._test_subsets
            second_test = self._test_complements
        else:
            first_test = self._test_complements
            second_test = self._test_subsets

        while True:
            assert self.test(config, 'assert') == self.FAIL

            subsets = self._split(config, n)

            logger.info('Run #%d: trying %s.', run, ' + '.join([str(len(subsets[i])) for i in range(n)]))

            next_config, next_n, complement_offset = first_test(n, run, config, subsets, complement_offset)
            if next_config is None:
                next_config, next_n, complement_offset = second_test(n, run, config, subsets, complement_offset)
            failed = next_config is not None

            if not failed:
                next_config = config
                next_n = min(len(config), n * 2)
                logger.info('Increase granularity to %d.', next_n)
                complement_offset = (complement_offset * next_n) / n

            # Minimization ends if no interesting configuration was found by the finest splitting or
            # if the configuration is already reduced to a single unit.
            if not failed and n == len(config):
                # No further minimizing
                logger.info('Done.')
                return config

            if failed and len(next_config) == 1:
                # No further minimizing
                logger.info('Done.')
                return next_config

            config = next_config
            n = next_n
            run += 1

    def _test_subsets(self, n, run, config, subsets, complement_offset):
        """
        Perform a subset based reduce task.

        :param n: The number of sets that the config is split to.
        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset need to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        for i in self._subset_iterator(n):
            if i is None:
                continue

            config_id = AbstractDD.config_id(run, 's', i)

            # Get the outcome either from cache or by testing it.
            outcome = self.lookup_cache(subsets[i], config_id) or self.test(subsets[i], config_id)
            if outcome == self.FAIL:
                # Interesting subset is found.
                logger.info('Reduced to %d units.', len(subsets[i]))
                logger.debug('New config: %r.', subsets[i])

                return subsets[i], 2, 0

        return None, None, complement_offset

    def _test_complements(self, n, run, config, subsets, complement_offset):
        """
        Perform a complement based reduce task.

        :param n: The number of sets that the config is split to.
        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset need to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        for j in self._complement_iterator(n):
            if j is None:
                continue

            i = int((j + complement_offset) % n)
            config_id = AbstractDD.config_id(run, 'c', i)
            complement = self.minus(config, subsets[i])

            outcome = self.lookup_cache(complement, config_id) or self.test(complement, config_id)
            if outcome == self.FAIL:
                # Interesting complement is found.
                logger.info('Reduced to %d units.', len(complement))
                logger.debug('New config: %r.', complement)

                # In next run, start removing the following subset
                return complement, max(n - 1, 2), i

        return None, None, complement_offset



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import argparse
import chardet
import codecs
import logging
import os
import pkgutil
import time

from os.path import abspath, basename, exists, join, relpath
from shutil import rmtree

from . import config_splitters, config_iterators, outcome_cache
from .combined_iterator import CombinedIterator
from .combined_parallel_dd import CombinedParallelDD
from .light_dd import LightDD
from .parallel_dd import ParallelDD
from .shared_cache import shared_cache_decorator
from .subprocess_test import ConcatTestBuilder, SubprocessTest

logger = logging.getLogger('picire')
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()


def create_parser():
    parser = argparse.ArgumentParser(description='Command line interface of the "picire" test case reducer')
    parser.add_argument('-i', '--input', metavar='FILE', required=True,
                        help='test case to be reduced')

    # Base reduce settings.
    parser.add_argument('--cache', metavar='NAME',
                        choices=[i for i in dir(outcome_cache) if not i.startswith('_') and i.islower()], default='config',
                        help='cache strategy (%(choices)s; default: %(default)s)')
    parser.add_argument('--split', metavar='NAME', dest='split_method',
                        choices=[i for i in dir(config_splitters) if not i.startswith('_')], default='zeller',
                        help='split algorithm (%(choices)s; default: %(default)s)')
    parser.add_argument('--test', metavar='FILE', required=True,
                        help='test command that decides about interestingness of an input')
    parser.add_argument('--encoding', metavar='NAME',
                        help='test case encoding (default: autodetect)')

    # Extra settings for parallel reduce.
    parser.add_argument('-p', '--parallel', action='store_true', default=False,
                        help='run DD in parallel')
    parser.add_argument('-c', '--combine-loops', action='store_true', default=False,
                        help='combine subset and complement check loops for more parallelization (has effect in parallel mode only)')
    parser.add_argument('-j', '--jobs', metavar='N', type=int, default=os.cpu_count(),
                        help='maximum number of test commands to execute in parallel (has effect in parallel mode only; default: %(default)s)')
    parser.add_argument('-u', '--max-utilization', metavar='N', type=int, default=100,
                        help='maximum CPU utilization allowed; don\'t start new parallel jobs until utilization is higher (has effect in parallel mode only; default: %(default)s)')

    # Tweaks how to walk through the chunk lists.
    parser.add_argument('--complement-first', dest='subset_first', action='store_false', default=True,
                        help='check complements first')
    parser.add_argument('--subset-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through subsets (%(choices)s; default: %(default)s)')
    parser.add_argument('--complement-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through complements (%(choices)s; default: %(default)s)')

    # Additional settings.
    parser.add_argument('-l', '--log-level', metavar='LEVEL',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'DISABLE'], default='INFO',
                        help='verbosity level of diagnostic messages (%(choices)s; default: %(default)s)')
    parser.add_argument('-v', '--verbose', dest='log_level', action='store_const', const='DEBUG', default=argparse.SUPPRESS,
                        help='verbose mode (alias for -l %(const)s)')
    parser.add_argument('-q', '--quiet', dest='log_level', action='store_const', const='DISABLE', default=argparse.SUPPRESS,
                        help='quiet mode (alias for -l %(const)s)')
    parser.add_argument('-o', '--out', metavar='DIR',
                        help='working directory (default: input.timestamp)')
    parser.add_argument('--disable-cleanup', dest='cleanup', default=True, action='store_false',
                        help='disable the removal of generated temporary files')
    return parser


def process_args(parser, args):
    if args.log_level == 'DISABLE':
        args.log_level = logging.CRITICAL + 1

    args.input = abspath(relpath(args.input))
    if not exists(args.input):
        parser.error('Input file does not exits: %s' % args.input)

    with open(args.input, 'rb') as f:
        args.src = f.read()

    if args.encoding:
        try:
            codecs.lookup(args.encoding)
        except LookupError:
            parser.error('The given encoding (%s) is not known.' % args.encoding)
    else:
        args.encoding = chardet.detect(args.src)['encoding']
        if not args.encoding:
            parser.error('The encoding of the test is not recognized.'
                         'Please define it with the --encoding command line option.')

    args.test = abspath(relpath(args.test))
    if not exists(args.test) or not os.access(args.test, os.X_OK):
        parser.error('Tester program does not exist or isn\'t executable: %s' % args.test)

    args.tester_class = SubprocessTest
    args.tester_config = {
        'encoding': args.encoding,
        'command_pattern': '%s %%s' % args.test
    }

    args.cache = getattr(outcome_cache, args.cache)
    if args.parallel:
        args.cache = shared_cache_decorator(args.cache)

    split_method = getattr(config_splitters, args.split_method)
    subset_iterator = getattr(config_iterators, args.subset_iterator)
    complement_iterator = getattr(config_iterators, args.complement_iterator)

    # Choose the reducer class that will be used and its configuration.
    args.reduce_config = {'split': split_method}
    if not args.parallel:
        args.reduce_class = LightDD
        args.reduce_config['subset_iterator'] = subset_iterator
        args.reduce_config['complement_iterator'] = complement_iterator
        args.reduce_config['subset_first'] = args.subset_first
    else:
        args.reduce_config['proc_num'] = args.jobs
        args.reduce_config['max_utilization'] = args.max_utilization

        if args.combine_loops:
            args.reduce_class = CombinedParallelDD
            args.reduce_config['config_iterator'] = CombinedIterator(args.subset_first, subset_iterator,
                                                                     complement_iterator)
        else:
            args.reduce_class = ParallelDD
            args.reduce_config['subset_iterator'] = subset_iterator
            args.reduce_config['complement_iterator'] = complement_iterator
            args.reduce_config['subset_first'] = args.subset_first

    args.out = abspath(relpath(args.out if args.out else '%s.%s' % (args.input, time.strftime('%Y%m%d_%H%M%S'))))


def call(*,
         reduce_class, reduce_config,
         tester_class, tester_config,
         input, src, encoding, out,
         atom,
         cache_class=None, cleanup=True):
    """
    Execute picire as if invoked from command line, however, control its
    behaviour not via command line arguments but function parameters.

    :param reduce_class: Reference to the reducer class.
    :param reduce_config: Dictionary containing information to initialize the reduce_class.
    :param tester_class: Reference to a runnable class that can decide about the interestingness of a test case.
    :param tester_config: Dictionary containing information to initialize the tester_class.
    :param input: Path to the test case to reduce (only used to determine the name of the output file).
    :param src: Contents of the test case to reduce.
    :param encoding: Encoding of the input test case.
    :param out: Path to the output directory.
    :param atom: Input granularity to work with during reduce ('char' or 'line').
    :param cache_class: Reference to the cache class to use.
    :param cleanup: Binary flag denoting whether removing auxiliary files at the end is enabled (default: True).
    :return: The path to the minimal test case.
    """

    # Get the parameters in a dictionary so that they can be pretty-printed
    # (minus src, as that parameter can be arbitrarily large)
    args = locals().copy()
    del args['src']
    logger.info('Reduce session starts for %s\n%s',
                input, ''.join(['\t%s: %s\n' % (k, v) for k, v in sorted(args.items())]))

    tests_dir = join(out, 'tests')
    os.makedirs(tests_dir, exist_ok=True)

    # Split source to the chosen atoms.
    if atom == 'line':
        content = src.decode(encoding).splitlines(keepends=True)
    elif atom == 'char':
        content = src.decode(encoding)
    logger.info('Initial test contains %d %ss', len(content), atom)

    test_builder = ConcatTestBuilder(content)
    cache = cache_class() if cache_class else None
    if hasattr(cache, 'set_test_builder'):
        cache.set_test_builder(test_builder)

    dd = reduce_class(tester_class(test_builder=test_builder,
                                   test_pattern=join(tests_dir, '%s', basename(input)),
                                   **tester_config),
                      cache=cache,
                      **reduce_config)
    min_set = dd.ddmin(list(range(len(content))))

    logger.debug('The cached results are: %s', cache)
    logger.debug('A minimal config is: %r', min_set)

    out_file = join(out, basename(input))
    with codecs.open(out_file, 'w', encoding=encoding, errors='ignore') as f:
        f.write(ConcatTestBuilder(content)(min_set))
    logger.info('Result is saved to %s.', out_file)

    if cleanup:
        rmtree(tests_dir)

    return out_file


def execute():
    """
    The main entry point of picire.
    """

    parser = create_parser()
    # Implementation specific CLI options that are not needed to be part of the core parser.
    parser.add_argument('-a', '--atom', metavar='NAME', choices=['char', 'line'], default='line',
                        help='atom (i.e., granularity) of input (%(choices)s; default: %(default)s)')
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))

    args = parser.parse_args()
    process_args(parser, args)

    logging.basicConfig(format='%(message)s')
    logger.setLevel(args.log_level)

    call(reduce_class=args.reduce_class,
         reduce_config=args.reduce_config,
         tester_class=args.tester_class,
         tester_config=args.tester_config,
         input=args.input,
         src=args.src,
         encoding=args.encoding,
         out=args.out,
         atom=args.atom,
         cache_class=args.cache,
         cleanup=args.cleanup)
